"""Compute an importance score for ACAI entities.

This is a simple heuristic based on the number of times an entity is
mentioned, the distribution of those mentions, and other type-specific
factors.

The importance score is a floating point number between 0.0 and 1.0,
where 1.0 means maximally important. Scores are normalized within
entity type, but it is explicitly not guaranteed that the same
importance score for two different entity types means the same
thing. For example, a score of 0.5 for a person may not be comparable
to a score of 0.5 for a place.

>>> from acai.core import importance
"""

from collections import UserDict
from pathlib import Path
from typing import Sized

import numpy as np
import pandas as pd

from acailib import DATAPATH

from acailib.shared_classes import AcaiEntity


class BaseImportance(UserDict):
    """Base class for importance scores.

    This class is not intended to be used directly. Instead, use one of
    the subclasses for a specific entity type.

    """

    # a list of functions and the weights they contribute to the
    # overall score. Weights must sum to 1.0.
    factors: dict[str, float] = {}
    # the type of entity this importance score is for, e.g. "people",
    type: str = "UNDEFINED"

    def __init__(self) -> None:
        """Initialize the importance scores for a set of entities."""
        super().__init__()

    def __post_init__(self) -> None:
        """Check values after initialization."""
        assert (
            sum(self.factors.values()) == 1.0
        ), f"Weights must sum to 1.0, not {sum(self.factors.values())}."
        self.basepath: Path = DATAPATH / "adjunct" / self.type
        self.basepath.mkdir(parents=True, exist_ok=True)

    @staticmethod
    def _decrement_length_weight(
        items: Sized,
        # base_weight: float = 1.0, decrement: float = 0.2
    ) -> float:
        """Decrement the length weight of a list of items.

        The more items there are, the less important each additional one is.
        """
        total = 0.0
        for i in range(len(items)):
            # weight = max(base_weight - i * decrement, 0.0)
            # try an exponential decrement
            # weight = max(base_weight - i * decrement, 0.0)
            total += 1 / (i + 1)
        return total

    @staticmethod
    def normalize_to_unit_range(arr: np.ndarray) -> np.ndarray:
        """Normalize a numpy array to the range [0, 1].

        This is done by subtracting the minimum value and dividing by
        the range (max - min). If all values are the same, return an
        array of zeros.

        """
        arr = np.asarray(arr)
        min_val = np.min(arr)
        max_val = np.max(arr)
        if max_val == min_val:
            return np.zeros_like(arr)
        return (arr - min_val) / (max_val - min_val)

    # these methods are common to most entity types

    # not sure this is actually useful
    def biblical(self, entry: AcaiEntity) -> float:
        """Return 1.0 if non_biblical is False, else 0.0."""
        assert hasattr(entry, "non_biblical"), f"Entry {entry} must have 'non_biblical' attribute."
        return 1.0 if entry.non_biblical is False else 0.0

    def is_nt(self, entry: AcaiEntity) -> float:
        """Return 1.0 if referenced in the New Testament, else 0.0."""
        assert hasattr(entry, "references"), f"Entry {entry!r} must have 'references' attribute."
        bookrefset: set[int] = {int(ref[:2]) for ref in entry.references}
        if bookrefset:
            return 1.0 if (max({int(ref[:2]) for ref in entry.references}) >= 40) else 0.0
        else:
            return 0.0

    # not sure this is actually useful
    def nonapochryphal(self, entry: AcaiEntity) -> float:
        """Return 1.0 if only_mentioned_in_apocrypha is False, else 0.0."""
        assert hasattr(
            entry, "only_mentioned_in_apocrypha"
        ), f"Entry {entry!r} must have 'only_mentioned_in_apocrypha' attribute."
        return 1.0 if entry.only_mentioned_in_apocrypha is False else 0.0

    # maybe NT references should be weighted higher than OT??

    def n_book_references(self, entry: AcaiEntity) -> int:
        """Compute the number of book references for a person."""
        # if entry.id not in self.is_primary:
        #     allids = self.aliases[self.aliases[entry.id][0]]
        aliased_entries = self._get_aliased_entries(entry)
        return len({ref[:2] for ent in aliased_entries for ref in ent.references})

    def n_chapter_references(self, entry: AcaiEntity) -> int:
        """Compute the number of chapter references for a person."""
        aliased_entries = self._get_aliased_entries(entry)
        return len({ref[:4] for ent in aliased_entries for ref in ent.references})

    def n_key_references(self, entry: AcaiEntity) -> int:
        """Return the number of key references for a person."""
        aliased_entries = self._get_aliased_entries(entry)
        return len({ref for ent in aliased_entries for ref in ent.key_references})

    def n_verse_references(self, entry: AcaiEntity) -> int:
        """Compute the number of verse references for a person."""
        aliased_entries = self._get_aliased_entries(entry)
        return len({ref for ent in aliased_entries for ref in ent.references})
        # maybe i don't need this complexity
        # def _bcv(ref):
        #     """Convert a word-level reference to a book-chapter-verse string."""
        #     bcv = ref.split(".")
        #     if re.match(r"^[n|o]", ref):
        #         bcv = bcv[1:]
        #     return bcv[:8]

        # return len(
        #     set(
        #         entry.references
        #         + [_bcv(ref) for reflist in entry.pronominal_referents for ref in reflist]
        #         + [_bcv(ref) for reflist in entry.subject_referents for ref in reflist]
        #     )
        # )

    def compute_importance(self, row: pd.Series) -> float:
        """Compute the importance score for an entity.

        This is done by summing the weighted factors for the entity.

        """
        score = 0.0
        for factor, weight in self.factors.items():
            score += getattr(row, factor) * weight
        return score

    def write_tsv(self) -> None:
        """Write the importance scores to a TSV file.

        The file will contain the entity ID and the importance score.

        """
        outpath = self.basepath / f"{self.type}_importance.tsv"
        # sort by importance score
        self.entitydf = self.entitydf.sort_values("importance", ascending=False)
        self.entitydf.to_csv(outpath, sep="\t", index=True, header=True)



"""The Annotator delivers ACAI entities for verses and words.

>>> from acai.core import Annotator
>>> ann = Annotator.Annotator()
# no annotations for the first word of MRK 1:4
>>> ann.get("n41001004001")
# second word is an explicit reference to John (the Baptist)
>>> ann.get("n41001004002")
[('person:John', 'explicit_instances')]
# fourth word has John as implicit subject
>>> ann.get("n41001004004")
[('person:John', 'subject_referents')]

"""

from collections import UserDict
from typing import Any

from biblelib.word import bcvwpid

from acai.util import EntityReader
from acai.images import Reader


# def imagereflookup(refstr: str) -> Any:
#     try:
#         # try parsing as USFM
#         bcvid = bcvwpid.fromusfm(refstr).get_id()
#         # usfm = refstr
#     except Exception as _:
#         # otherwise treat as BCV
#         bcvid = refstr
#         # usfm = bcvwpid.make_id(refstr).to_usfm()
#     return set(READER.get_images(bcvid))


class Annotator(UserDict):

    reader = Reader()

    readers: dict[str, EntityReader.EntityReader] = {
        "deity": EntityReader.DeitiesReader(),
        "group": EntityReader.GroupsReader(),
        "person": EntityReader.PeopleReader(),
        "place": EntityReader.PlacesReader(),
    }

    # ntrefs = {ref: data for ref, data in ann.items() if int(ref[1:3]) > 39}
    def __init__(self) -> None:
        """Initialize an instance."""
        super().__init__(self)
        # self.data maps BCVWP identifiers to a list of (entityid,
        # referencetype) tuples
        # data: dict[str, list[tuple[str, str]]] = {}
        for readertype, reader in self.readers.items():
            for entityid, entityrec in reader.items():
                entityrefdict = reader.word_references_by_type(entityid)
                for reftype in entityrefdict:
                    # this 'flattens' sequences of tokens: that may be wrong
                    for canon, allrefs in entityrefdict[reftype].items():
                        for reflist in allrefs:
                            for ref in reflist:
                                if ref not in self.data:
                                    self.data[ref] = []
                                self.data[ref].append((entityid, reftype))
        # also need to capture ref -> image, but this is verse-level,
        # not word-level (until associated with ACAI entities)



"""Code for working with ACAI data."""

from enum import Enum
import os
from pathlib import Path

# the dotenv magic only works if your Python environment is running in
# the right directory
import dotenv

# use an environment variable if available
if not dotenv.load_dotenv():
    print("No .env file found")

# the root of Clear-Bible repositories
clearrootenvar = os.getenv("CLEARROOT")
if clearrootenvar:
    CLEARROOT = Path(clearrootenvar)
else:
    # guess if not defined in .env
    CLEARROOT = Path.home() / "git/Clear-Bible"
    print(f"No environment variable for CLEARROOT: assuming {CLEARROOT}")

# the root of a local copy of BibleAquifer/ACAI
# Currently you need this for loading data
acairootenvar = os.getenv("ACAIROOT")
if acairootenvar:
    ACAIROOT = Path(acairootenvar)
else:
    # guess if not defined in .env
    ACAIROOT = Path.home() / "git/BibleAquifer/ACAI"
    print(f"No environment variable for ACAIROOT: assuming {ACAIROOT}")


ROOT = Path(__file__).parent.parent
DATAPATH = ROOT / "data"


OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")


class LangEnum(str, Enum):
    """Enumerate lang code values for biblical source languages."""

    aramaic = "arc"
    greek = "grc"
    hebrew = "hbo"


class TargetLanguageCodes(str, Enum):
    """Enumerate ISO-639-3 lang code values for target LWC languages."""

    ARABIC = "arb"
    BISLAMA = "bis"
    ENGLISH = "eng"
    FRENCH = "fra"
    HAUSA = "hau"
    HINDI = "hin"
    IGBO = "ibo"
    INDONESIAN = "ind"
    # Ethnologue lists several: is this the right one?
    MALAY = "pse"
    # not sure how to represent Traditional
    MANDARIN = "cmn"
    PORTUGUESE = "por"
    RUSSIAN = "rus"
    SPANISH = "spa"
    SWAHILI = "swa"
    TOK_PISIN = "tpi"
    VIETNAMESE = "vie"


class ContentType(str, Enum):
    """Enumerate types for images and other assets."""

    # from UBS Image data
    ILL = "illustration"
    IMG = "image"
    MAP = "map"
    # extensions
    PHOTO = "photo"


__all__ = [
    "CLEARROOT",
    "ACAIROOT",
    "ROOT",
    "DATAPATH",
    "OPENAI_API_KEY",
    "LangEnum",
    "TargetLanguageCodes",
    "ContentType",
]



"""Utilities for working with ACAI data."""

import re

from .TSVReader import TSVReader
from .Language import Locstr, LocDict, LocDictSet, LanguageSet
from .marble import is_prot_marbleid, from_marbleid, from_marbleid_range, get_from_marbleids

from .strongs import normalize_strongs


__all__ = [
    #
    "makeuri",
    # Language
    "Locstr",
    "LocDict",
    "LocDictSet",
    "LanguageSet",
    # TSVReader
    "TSVReader",
    # marble
    "is_prot_marbleid",
    "from_marbleid",
    "from_marbleid_range",
    "get_from_marbleids",
    # strongs
    "normalize_strongs",
]


def makeuri(prefLabel: str, namespaceprefix: str) -> str:
    """Return a URI for an ACAI item based on the preferred label.

    This replaces and strips characters from the title to make a URI.
    """
    assert prefLabel, "Empty prefLabel is not allowed."
    assert namespaceprefix.endswith(":"), f"'{namespaceprefix}' must end in a colon"
    uri = re.sub(r"[ ~/-]", "_", prefLabel)
    cleanuri = re.sub(r"[“”’():]", "", uri)
    return f"{namespaceprefix}{cleanuri}"



"""Categorize according to content type, based on name matching.


>>> from acai.util import TyndaleReader
>>> tynrd = TyndaleReader.Reader()
>>> len(tynrd)
5027
>>> list(tynrd.items())[23]
('AbelShittim', Article(identifier='AbelShittim', headword='ABEL-SHITTIM', paragraphs=['Alternate name for Shittim, a place on\xa0the plains of Moab, in bref^Num_33_49. See art^ShittimPlace.'], wordcount=14, isredirect=True, contenttype=''))
>>> from acai.util import Categorizer
>>> ctgr = Categorizer.Categorizer()
# fil in the contenttype attribute
>>> for art in tynrd.values():
...   art.contenttype = ctgr.categorize_tynbd(art.clean_headword)
...
>>> tynrd["AbelShittim"]
Article(identifier='AbelShittim', headword='ABEL-SHITTIM', paragraphs=['Alternate name for Shittim, a place on\xa0the plains of Moab, in bref^Num_33_49. See art^ShittimPlace.'], wordcount=14, isredirect=True, contenttype='place')

"""

from collections import defaultdict

from acai.util import EntityReader


class Categorizer:
    """Match names to heuristically determine content types."""

    default_type: str = "acai"
    upcase: bool = True
    entitytypes: list = ["deity", "group", "person", "place"]

    def __init__(self) -> None:
        """Initialize an instance."""
        self.readers: dict[str, EntityReader.EntityReader] = {
            "deity": EntityReader.DeitiesReader(self.upcase),
            "group": EntityReader.GroupsReader(self.upcase),
            "person": EntityReader.PeopleReader(self.upcase),
            "place": EntityReader.PlacesReader(self.upcase),
        }

    @staticmethod
    def clean_asterisk(string: str) -> str:
        """Drop a final asterisk if present."""
        if string.endswith("*"):
            return string[:-1]
        else:
            return string

    def categorize_string(self, string: str) -> str:
        """Return a content type string given a Tyndale headword.

        This uses a sequence of heuristics: the categorization is not
        guaranteed to be correct.

        If nothing matches, return the default_type.

        """
        # special cases for TynBD headwords
        if string.endswith(" (Person )"):
            return "person"
        elif string.endswith(" (Place )"):
            return "place"
        elif string in self.readers["person"].eng_preferred_label_dict:
            return "person"
        elif string in self.readers["place"].eng_preferred_label_dict:
            return "place"
        elif string in self.readers["group"].eng_preferred_label_dict:
            return "group"
        elif string in self.readers["deity"].eng_preferred_label_dict:
            return "deity"
        elif string in self.readers["person"].eng_alternate_label_dict:
            return "person"
        elif string in self.readers["place"].eng_alternate_label_dict:
            return "place"
        else:
            return self.default_type

    def categorize_tynbd(self, headword: str) -> str:
        """Return a content type string given a Tyndale headword.

        This uses a sequence of heuristics: the categorization is not
        guaranteed to be correct.

        If nothing matches, return the default_type.

        """
        if ", " in headword:
            types: list[str] = [
                self.categorize_string(self.clean_asterisk(string))
                for string in headword.split(", ")
            ]
            realtypes = [t for t in types if t != self.default_type]
            if realtypes:
                return realtypes[0]
            else:
                return self.default_type
        else:
            return self.categorize_string(self.clean_asterisk(headword))


"""Incremental development:

- Simple entity name matching gets 2105 categories
- Adding headword cleanup for trailing asterisk: 2575
- clean_headword splits commas: 2868
- adding checking of alternate labels: 3225

Additional patterns not yet implemented:

- comma inversions like "ARABAH*, Brook of the"
- books: variety of string matches, 66+
- art.headword.endswith("Tribe of"): 12
- Hebrew months: 12
- Gates
"""



"""Common Pydantic classes for language support.

>>> from acai import LangEnum
>>> from acai.util import Language
>>> lstr = Language.Locstr(lang=LangEnum("grc"), value="foo")
>>> lstr.lang
<LangEnum.greek: 'grc'>
>>> lstr.lang.value
'grc'
>>> lset = Language.LanguageSet(lang="grc", lemma="Ἰησοῦς", transliteration="Iēsŏus", references=["n41001001004"])
>>> lset.lang
<LangEnum.greek: 'grc'>
>>> lset.lemma
'Ἰησοῦς'
>>> ld = Language.LocDict({"eng": "dog", "spa": "perro"})
>>> ld.get_value("eng")
'dog'
# language code must exist and have data
>>> ld.get_value("fra")
AssertionError: No data for fra


For ingestion, it may be best to collate values from several fields
together for a single language, so you can e.g. retrieve all the
English values without worrying about the others.

PROBLEM: switching from Pydantic to ordinary dataclasses because of a
problem with downstream Pydantic classes using them.

"""

from collections import UserDict
from dataclasses import dataclass
from typing import Any, Mapping

# from pydantic import BaseModel

from acailib import LangEnum, TargetLanguageCodes


# class Locstr(BaseModel):
@dataclass
class Locstr:
    """Defines an instance of a localized string for a biblical language."""

    # string value
    value: str
    # language code
    lang: LangEnum = LangEnum.greek

    def __post_init__(self) -> None:
        """Compute values after initialization."""
        assert isinstance(self.lang, LangEnum), f"lang must be a LangEnum instance: {self.lang}"


class _LocBase(UserDict):
    """Base for collections of localized strings."""

    targetlangcodes = {tlc.value for tlc in TargetLanguageCodes.__members__.values()}

    def __init__(self, initialdata: Mapping[str, Any] = {}) -> None:
        """Initialize a LocDict instance.

        Checks the validity of keys, but not values.
        """
        for pair in initialdata.items():
            self._check_mapping(*pair)
        super().__init__(initialdata)

    def _check_mapping(self, code: str, value: Any) -> None:
        """Check that mapping is valid: return False if not."""
        self._check_langcode(code)
        self._check_value(value)

    def _check_langcode(self, code: str) -> None:
        """Check that language code is valid: return False if not."""
        assert code in self.targetlangcodes, f"Invalid language code: {code}"

    def _check_value(self, value: Any) -> None:
        """Check to ensure value is valid."""
        assert isinstance(value, str), f"Value must be a string: {value}"

    # typing complaints here about incompatible signatures: beyond my
    # understanding
    def update(self, m: Mapping[str, str]) -> None:
        """Update, with key checking."""
        # unlike parent, doesn't allow empty m
        for code, val in m.items():
            self._check_mapping(code, val)
            self[code] = val

    def get_value(self, code: Any) -> Any:
        """Return the value for code, a language code.

        It is an error if there is no localization for code.
        """
        self._check_langcode(code)
        assert code in self, f"No data for {code}"
        return self[code]


class LocDict(_LocBase):
    """Defines a collection of localized strings.

    Assumes the various strings are translations of each
    other. Languages must either be valid source or target codes.

    """

    def __init__(self, initialdata: Mapping[str, str]) -> None:
        """Initialize a LocDict instance."""
        super().__init__(initialdata)


class LocDictSet(_LocBase):
    """Defines a collection of localizations, with a set for each language.

    This is typically used for alternate labels. There is no
    correspondence between labels across languages.

    """

    def __init__(self, initialdata: Mapping[str, set]) -> None:
        """Initialize a LocDict instance."""
        super().__init__(initialdata)

    # override
    def _check_value(self, value: set) -> None:
        """Check to ensure value is valid.

        Value must be a non-empty set.
        """
        assert isinstance(value, set), f"Value must be a set: {value}"
        assert value, "Set must not be empty"


# class LanguageSet(BaseModel):
@dataclass
class LanguageSet:
    """Defines a lemma and references for a dataset instance."""

    # language code
    lang: LangEnum
    lemma: str
    transliteration: str
    # list of BCVWP IDs
    # must be at least one: needs validation
    references: list[str]

    def __post_init__(self) -> None:
        """Compute values after initialization."""
        assert isinstance(self.lang, LangEnum), f"lang must be a LangEnum instance: {self.lang}"



"""Serialize UBS DGNT data to files for other uses.

>>> from acai.lexicon import serializer
>>> serializer.Serializer(outpath=(ROOT.parent / "TyndaleBibleDictionary/notebooks/ubsdgnt.tsv"))
"""

from csv import DictWriter
from pathlib import Path

from . import ubsdgnt, ROOT


class Serializer:
    """Serialize UBS DGNT data to files for other uses.

    This is a simple serializer that writes the data to a TSV file.
    """

    ubsprs = ubsdgnt.UBSParser()
    fieldnames = ["entryCode", "lemma", "glosses", "definitionShort", "lexDomains", "lexSubDomains"]

    def __init__(self, outpath: Path) -> None:
        """Initialize the serializer."""
        with outpath.open("w", encoding="utf-8", newline="") as f:
            writer = DictWriter(f, fieldnames=self.fieldnames, delimiter="\t")
            writer.writeheader()
            for meaning in self.ubsprs.meaning_iterator():
                _ = writer.writerow(meaning.serialize_text())



"""Load external license data.

Not much validation required since this is a simple input file.

>>> from acai.util import license
>>> rd = license.TSVReader
>>> rd["CC0"]
License(license_id='CC0', license_url='https://creativecommons.org/publicdomain/zero/1.0/',
  attribution=False, sharealike=False, noderivs=False, noncommercial=False)
>>> [lic.license_id for lic in rd.values() if not lic.noncommercial]
['CC0', 'CC BY 1.0', 'CC BY-SA 1.0', 'CC BY-ND 1.0', ...]

"""

from pathlib import Path

from pydantic import BaseModel  # type: ignore

from acai import DATAPATH
from .TSVReader import TSVReader as tsvr


class License(BaseModel):
    """Manage data that is read in from file on a license.

    This should stay in sync with license.tsv.
    """

    license_id: str
    license_url: str
    attribution: bool
    sharealike: bool
    noderivs: bool
    noncommercial: bool


class TSVReader(tsvr):
    """Read TSV data into a dict.

    This is generalized for use where the source data is in TSV, with
    column headers that match a BaseModel instance, and a designated
    identity element.

    """

    # subs must define these
    tsvpath: Path = DATAPATH / "license.tsv"
    idattr: str = "license_id"
    model: BaseModel = License



"""Aggregate references into an index."""

import acailib.EntityReader as er


class ReferencesIndex:
    """Index of BCV references.

    This is multi-level: reference -> reftype -> list of entity IDs.

    Example: "01041050" -> {'key_references': ['person:Manasseh', 'person:Joseph.10', ...],}
    for 'person:Manasseh',

    Caveats:
    - 'references' may include metonymies (especially for the patriarchs of the 12 tribes)
    - 'key_references' may not be named references
    """



"""Generalized Reader for TSV data.

Subclass this, defining tsvpath, idattr, and model.
"""

from collections import UserDict
from csv import DictReader
from pathlib import Path
from typing import Any


class TSVReader(UserDict):
    """Read TSV data into a dict.

    This is generalized for use where the source data is in TSV, with
    column headers that match a BaseModel instance, and a designated
    identity element.

    """

    # subs must define these
    tsvpath: Path
    idattr: str
    model: Any

    def __init__(self) -> None:
        """Initialize a Reader instance."""
        super().__init__()
        assert self.tsvpath, "tsvpath must be defined"
        assert self.idattr, "idattr must be defined"
        assert self.model, "model must be defined"
        with self.tsvpath.open() as f:
            reader = DictReader(f, delimiter="\t")
            self.data = {
                idattr: self.model(**row) for row in reader if (idattr := row[self.idattr])
            }



"""Read ACAI data into a dict.

>>> import acailib.EntityReader as er
>>> peoplerd = er.PeopleReader()
# return a list of entities whose preferred label matches a given string
>>> peoplerd.eng_preferred_label_dict["John"]
['person:John.3', 'person:John', 'person:John.2', 'person:John.5', 'person:John.4']
# likewise for other EntityReader subclasses

"""

from collections import UserDict, defaultdict
from collections.abc import Callable
import json
from pathlib import Path
from typing import Optional
from warnings import warn

from acailib import ACAIROOT
from acailib.shared_classes import AcaiDeity, AcaiGroup, AcaiPerson, AcaiPlace


class EntityReader(UserDict):
    """Base class for reading entity data."""

    acaidir: str = ""
    jsondir: Path = Path()
    # typing here needs help
    entryclass: Optional[Callable] = None
    repodata: str = "https://github.com/Clear-Bible/ACAI/tree/main/data"

    def __init__(self, upcase: bool = False) -> None:
        """Initialize an instance.

        With upcase = True (default is False), convert names to all
        upper-case. This is useful for matching TynBD headwords that
        are all upper-case.

        """
        super().__init__()
        assert self.jsondir, "Need to instantiate a subclass of EntityReader"
        for jsonfile in self.jsondir.glob("*.json"):
            with jsonfile.open() as f:
                jdict = json.load(f)
            try:
                if self.entryclass:
                    self.data[jdict["id"]] = self.entryclass(**jdict)
            except Exception as e:
                print(f"Failed on {jsonfile}\n{e}")
        # these are all specific to an entity type for the subclass
        self.eng_preferred_label_dict: defaultdict = defaultdict(list)
        self.eng_alternate_label_dict: defaultdict = defaultdict(list)
        # this is to test for preflabel uniqness
        self._eng_preflabels: set = set()
        self._eng_preflabel_duplicates: defaultdict = defaultdict(list)
        # does not include `key_references`, `explicit_instances`,
        # `pronominal_referents`, or `subject_referents`, so use with
        # caution
        self.references_dict: defaultdict = defaultdict(list)
        for entityid in self.data:
            fullpreflabel: str = self[entityid].localizations["eng"]["preferred_label"]
            # currently about 130 of these that need resolution: should be none
            if fullpreflabel in self._eng_preflabels:
                warn(f"Duplicate preferred label {fullpreflabel} for {entityid}")
                self._eng_preflabel_duplicates[fullpreflabel].append(entityid)
            else:
                self._eng_preflabels.add(fullpreflabel)
            # split is a hack to drop post-name qualifiers
            preflabel = self[entityid].localizations["eng"]["preferred_label"].split(" ")[0]
            altlabels = self[entityid].localizations["eng"].get("alternate_labels", [])
            if upcase:
                preflabel = preflabel.upper()
                altlabels = [label.upper() for label in altlabels]
            self.eng_preferred_label_dict[preflabel].append(entityid)
            for label in altlabels:
                self.eng_alternate_label_dict[label].append(entityid)
            for ref in self[entityid].references:
                self.references_dict[ref].append(entityid)
            # ensure key references are all in the full references list
            # NO: sometimes key references are not in the full references list (e.g., for groups)
            # for keyref in self[entityid].key_references:
            #     if keyref not in self[entityid].references:
            #         warn(f"Key reference {keyref} not in full references list for {entityid}")

    def acai_uri(self, identifier: str) -> str:
        """Return a string for the Github repo location for identifier."""
        nsp, name = identifier.split(":")
        return f"{self.repodata}/{self.acaidir}/json/acai/{name}.json"

    def acai_md_link(self, identifier: str) -> str:
        """Return a Markdown link to the JSON file for identifier."""
        uri = self.acai_uri(identifier)
        return f"[{identifier}]({uri})"

    def references_by_type(self, identifier: str) -> dict[str, list[str]]:
        """Return all the references for an entity, grouped by type."""
        reftypes = [
            # note these overlap with other reference types
            "key_references",
            "references",
            "explicit_instances",
            "pronominal_referents",
            "subject_referents",
        ]
        return {reftype: getattr(self[identifier], reftype) for reftype in reftypes}

    def word_references_by_type(self, identifier: str) -> dict[str, list[str]]:
        """Return all the word-level references for an entity, grouped by type."""
        bcvwpreftypes = [
            "explicit_instances",
            "pronominal_referents",
            "subject_referents",
        ]
        return {
            reftype: references
            for reftype, references in self.references_by_type(identifier).items()
            if reftype in bcvwpreftypes
        }

    def preflabel(self, entityid: str, language: str = "eng") -> str:
        """Return the preferred label for ann entity in the specified language.

        Caveats:
        - Not all preflabels are actually names: occasionally they are
          descriptive terms (e.g. 'Aramitess' for Manasseh's unnamed
          concubine).

        """
        if language != "eng":
            raise NotImplementedError("Not implemented for languages other than English.")
        preflabel: str = self.data[entityid].preflabel(language)
        return preflabel


class DeitiesReader(EntityReader):
    """Read Deities data."""

    acaidir = "deities"
    jsondir = ACAIROOT / acaidir / "json"
    entryclass = AcaiDeity


class GroupsReader(EntityReader):
    """Read Groups data."""

    acaidir = "groups"
    jsondir = ACAIROOT / acaidir / "json"
    entryclass = AcaiGroup


class PeopleReader(EntityReader):
    """Read People data."""

    acaidir = "people"
    jsondir = ACAIROOT / acaidir / "json"
    entryclass = AcaiPerson

    def children_str(self, personid: str, language: str = "eng") -> str:
        """Return a string identifying children."""
        preflabel = self.preflabel(personid, language)
        personinst: AcaiPerson = self.data[personid]
        if personinst.offspring:
            childstr: str = ", ".join(
                [self.data[child].preflabel(language) for child in personinst.offspring]
            )
            if len(childstr) > 1:
                childrenstr = f"children of {preflabel} were"
            else:
                childrenstr = f"child of {preflabel} was"
            return f"The {childrenstr} {childstr}."
        else:
            return ""

    def partner_str(self, personid: str, language: str = "eng") -> str:
        """Return a string identifying partners."""
        preflabel = self.preflabel(personid, language)
        personinst: AcaiPerson = self.data[personid]
        if personinst.partners:
            spousestr: str = ", ".join(
                [self.data[partner].preflabel(language) for partner in personinst.partners]
            )
            if len(spousestr) > 1:
                partnerstr = f"partners of {preflabel} were"
            else:
                partnerstr = f"partner of {preflabel} was"
            return f"The {partnerstr} {spousestr}."
        else:
            return ""


class PlacesReader(EntityReader):
    """Read Places data."""

    acaidir = "places"
    jsondir = ACAIROOT / acaidir / "json"
    entryclass = AcaiPlace


# class FaunaReader(EntityReader):
#     """Read Places data."""

#     acaidir = "fauna"
#     jsondir = ACAIROOT / acaidir / "json"
#     entryclass = AcaiFaunaEntry



"""Utilities for working with ACAI Lexicon data."""

from acai import ROOT, DATAPATH

# from .ubsgnt import GNTReader
from .LexiconEntry import LEXSense, LEXMeaning, BaseForm, Lexicon_Entry
from .serializer import Serializer
from .ubsdgnt import UBSParser, Aggregator

__all__ = [
    "ROOT",
    "DATAPATH",
    # LexiconEntry
    "LEXSense",
    "LEXMeaning",
    "BaseForm",
    "Lexicon_Entry",
    # serializer
    "Serializer",
    # ubsdgnt
    "UBSParser",
    "Aggregator",
]



"""Use SpaCy to lemmatize English words."""

from collections import UserDict
from csv import DictWriter
from dataclasses import dataclass
from pathlib import Path

import spacy

from acai import DATAPATH
from acai.util import TSVReader


BBEWORDS = DATAPATH / "lexicon/uniqueWordsBBE.tsv"


@dataclass
class LemmatizedWord:
    """Represents a word and its lemmatized form."""

    word: str
    lemma: str = ""

    def diff_lemma(self) -> bool:
        """Return True if lemma is diferent from word."""
        return not self.word == self.lemma


class WordReader(TSVReader):
    """Read BBE words."""

    tsvpath = BBEWORDS
    # identity element from the TSV
    idattr = "word"
    # model to represent the data from each row
    model = LemmatizedWord


class Lemmatizer(UserDict):
    """Lemmatize English words."""

    # reads the data
    words = WordReader()
    nlp = spacy.load("en_core_web_sm")

    def __init__(self) -> None:
        """Initialize an instance."""
        super().__init__(self)
        for word, lword in self.words.items():
            # this does _all_ the analysis: speed it up by omitting
            # parts of the pipeline?
            doc = self.nlp(word)
            lword.lemma = doc[0].lemma_
        # mapping from lowercased word to true-case: for correction of
        # lemmatized forms
        self.lcasewords = {word.lower(): word for word in self.words.keys()}
        # map lemmatized forms to a list of their words
        for lword in self.words.values():
            if lword.lemma not in self.data:
                self.data[lword.lemma] = set()
            self.data[lword.lemma].add(lword.word)

    def write(self, outpath: Path) -> None:
        """Write out lemmatized forms."""
        fieldnames = ["lemma", "words"]
        with outpath.open("w") as f:
            writer = DictWriter(f, fieldnames=fieldnames, delimiter="\t")
            writer.writeheader()
            for lemma, wordset in self.items():
                # capitalize the lemma if there's a case-insensitive
                # matched word
                caplemma = self.lcasewords.get(lemma, lemma)
                writer.writerow({"lemma": caplemma, "words": "; ".join(wordset)})
